#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Colour.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/Obj.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"
#include "ngl/SpotLight.h"
#include <QString>


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  // now set the inital camera values
  m_cameraIndex=0;
  m_moveMode=MOVEEYE;
  m_drawHelp=true;
  m_fov=65.0;
  m_aspect=1024.0/768.0;
  m_projection=ngl::PERSPECTIVE;
	m_near=1.0;
	m_far=10.0;

}
void GLWindow::createCameras()
{
  // create a load of cameras
  ngl::Camera cam;
  ngl::Camera Tcam;
  ngl::Camera Scam;
  ngl::Camera Fcam;
  // set the different vectors for the camera positions
  ngl::Vector EYE(0,0,1);
  ngl::Vector LOOK=0.0;
  ngl::Vector UP(0,1,0);

  ngl::Vector TEYE(0,50,0);
  ngl::Vector TLOOK=0.0;
  ngl::Vector TUP(0,0,1);

  // finally set the cameras shape and position
  cam.set(EYE,LOOK,UP);
  cam.setShape(m_fov,m_aspect, m_near,m_far,m_projection);
  m_cameras.push_back(cam);
  Tcam.set(TEYE,TLOOK,TUP);
  Tcam.setShape(80,m_aspect, 0.5,100,m_projection);
  m_cameras.push_back(Tcam);
}
// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
  glEnable(GL_DEPTH_TEST);
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  #ifdef WIN32
  	glewInit();
  #endif
  Init->initGlew();
  createCameras();
  // create an instance of the VBO primitive
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  // create a plane
  prim->createVBOSphere("sphere",1.0,20);



  // now to load the shader and set the values
  // grab an instance of shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // load a frag and vert shaders for colour no shading
  shader->loadShader("Colour","shaders/Colour.vs","shaders/Colour.fs");
  // set this as the active shader
  shader->useShader("Colour");
  shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
	shader->setShaderParam4f("Colour","Colour",1,1,1,1);
  // load a frag and vert shaders
  shader->loadShader("Blinn","shaders/Vertex.vs","shaders/Fragment.fs");
  // set this as the active shader
  shader->useShader("Blinn");
  // now pass the modelView and projection values to the shader
  shader->setShaderParamFromMatrix("Blinn","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cameras[m_cameraIndex].getProjection());

  // now set the material and light values
  ngl::Material m(ngl::GOLD);
  m.use();
  ngl::Light L1(ngl::Vector(-1,1,0),ngl::Colour(1,1,1,1),ngl::LIGHTREMOTE);
	L1.enable();
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  // set the viewport
  glViewport(0,0,_w,_h);
  // calculate the aspect ratio
  m_aspect=(float)_w/_h;
  // now set the camera values
  setCameraShape();
 }

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // Rotation based on the mouse position for our global
  // transform
  ngl::Transformation trans;
  if(m_cameraIndex == 1)
	{
		trans.setRotation(m_spinXFace,m_spinYFace,0);
	}
	// set this in the TX stack
  m_transformStack.setGlobal(trans);
  // now set this value in the shader for the current Camera, we do all 3 elements as they
  // can all change per frame
  shader->useShader("Blinn");
  shader->setShaderParamFromMatrix("Blinn","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cameras[m_cameraIndex].getProjection());


  // get the VBO instance and draw the built in teapot
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();


  shader->setShaderParamFromMatrix("Blinn","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());

  shader->useShader("Blinn");
  // push a transform
	float ext=150.0;
	int numSpheresDrawn=0;
	int totalSpheres=0;
	ngl::Vector p;

	glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  for(float z=-ext; z<ext; z+=2)
	{
		for(float y=-ext; y<ext; y+=2)
		{
			for(float x=-ext; x<ext; x+=2)
			{
				++totalSpheres;
				p.set(x,y,z,1.0);
				if( m_cameras[0].isSphereInFrustum(p,1.0f) !=ngl::OUTSIDE )
				{
					++numSpheresDrawn;
					m_transformStack.pushTransform();
					{
						// translate into position
						m_transformStack.getCurrentTransform().setPosition(x,y,z);
						// pass values to the shader
						shader->setShaderParamFromMatrix("Blinn","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
						// draw
						prim->draw("sphere");
					}
					m_transformStack.popTransform();
				}
			}
		}
	}
	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

	shader->use("Colour");
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
  shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cameras[m_cameraIndex].getModelView());

	m_transformStack.pushTransform();
  {
		m_transformStack.loadGlobalAndCurrentMatrixToShader("Colour","ModelMatrix");
		m_cameras[0].drawFrustum();
	}
  m_transformStack.popTransform();
	if( m_cameraIndex ==1)
	{
		m_transformStack.pushTransform();
		{
			m_transformStack.getCurrentTransform().setPosition(m_cameras[0].getEye());
			m_transformStack.loadGlobalAndCurrentMatrixToShader("Colour","ModelMatrix");
			prim->draw("cube");
		}
		m_transformStack.popTransform();
	}
	// now we are going to draw the help
  if (m_drawHelp==true)
  {
   // first we need to fallback to the fixed functionality GL pipeline
   shader->useNullProgram();
	 // create a font for drawing
   QFont font;
   // set it to bold
   font.setBold(true);
   // an make it 16 point
   font.setPointSize(16);
   // now we use the deprecated glColour command to set the colour for the text
   // this will not work with 3/4 core profile but neither will Qt at present
   glColor3f(1,1,1);
   // now render the text using the QT renderText helper function
   QString text=QString("Total Spheres %1 num drawn %2").arg(totalSpheres).arg(numSpheresDrawn);
	 renderText(10,18,text,font);
	 text=QString("near %1 far %2").arg(m_near).arg(m_far);
	 renderText(10,40,text,font);
	 text=QString("fov %1").arg(m_fov);
	 renderText(10,60,text,font);
  }

}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}


void GLWindow::setCameraShape()
{
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  m_cameras[0].setShape(m_fov,m_aspect,m_near,m_far,m_projection);

  shader->useShader("Blinn");
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
  shader->useShader("Colour");
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
	updateGL();
	m_cameras[0].calculateFrustum();
}

void GLWindow::keyPress(
                        QKeyEvent *_event
                       )
{
  const static float keyIncrement=0.2;
  // this method is called every time the main window recives a key event.
  // we then switch on the key value and set the camera in the GLWindow
  // find which key and do appropriate event
  switch (_event->key())
  {
  case Qt::Key_Escape : { QApplication::exit(EXIT_SUCCESS); break; }
  case Qt::Key_1 : { m_cameraIndex=0; break; }
  case Qt::Key_2 : { m_cameraIndex=1; break; }
	case Qt::Key_3 : { updateNearFar(DECNEAR); break; }
	case Qt::Key_4 : { updateNearFar(INCNEAR); break; }
	case Qt::Key_5 : { updateNearFar(DECFAR); break; }
	case Qt::Key_6 : { updateNearFar(INCFAR); break; }

  case Qt::Key_E : { m_moveMode=MOVEEYE; break; }
  case Qt::Key_L : { m_moveMode=MOVELOOK; break; }
  case Qt::Key_B : { m_moveMode=MOVEBOTH; break; }
  case Qt::Key_S : { m_moveMode=MOVESLIDE; break; }
  case Qt::Key_H : { m_drawHelp^=true; break; }
  case Qt::Key_Plus : { ++m_fov; setCameraShape(); break; }
  case Qt::Key_Minus :{ --m_fov; setCameraShape(); break; }
  case Qt::Key_R : { m_cameras[0].roll(3);  break; }
  case Qt::Key_P : { m_cameras[0].pitch(3); break; }
  case Qt::Key_Y : { m_cameras[0].yaw(3);   break; }

  case Qt::Key_Left :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[0].moveEye(keyIncrement,0,0);  break;}
      case MOVELOOK :  { m_cameras[0].moveLook(keyIncrement,0,0); break;}
      case MOVEBOTH :  { m_cameras[0].moveBoth(keyIncrement,0,0); break;}
      case MOVESLIDE : { m_cameras[0].slide(keyIncrement,0,0);    break;}
    }
  break;
  } // end move left
  case Qt::Key_Right :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[0].moveEye(-keyIncrement,0,0);  break;}
      case MOVELOOK :  { m_cameras[0].moveLook(-keyIncrement,0,0); break;}
      case MOVEBOTH :  { m_cameras[0].moveBoth(-keyIncrement,0,0); break;}
      case MOVESLIDE : { m_cameras[0].slide(-keyIncrement,0,0);    break;}
    }
  break;
  } // end move right
  case Qt::Key_Up :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[0].moveEye(0,keyIncrement,0);  break;}
      case MOVELOOK :  { m_cameras[0].moveLook(0,keyIncrement,0); break;}
      case MOVEBOTH :  { m_cameras[0].moveBoth(0,keyIncrement,0); break;}
      case MOVESLIDE : { m_cameras[0].slide(0,keyIncrement,0);    break;}
    }
  break;
  } // end move up
  case Qt::Key_Down :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[0].moveEye(0,-keyIncrement,0);  break;}
      case MOVELOOK :  { m_cameras[0].moveLook(0,-keyIncrement,0); break;}
      case MOVEBOTH :  { m_cameras[0].moveBoth(0,-keyIncrement,0); break;}
      case MOVESLIDE : { m_cameras[0].slide(0,-keyIncrement,0);    break;}
    }
  break;
  } // end move down
  case Qt::Key_O:
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[0].moveEye(0,0,keyIncrement);  break;}
      case MOVELOOK :  { m_cameras[0].moveLook(0,0,keyIncrement); break;}
      case MOVEBOTH :  { m_cameras[0].moveBoth(0,0,keyIncrement); break;}
      case MOVESLIDE : { m_cameras[0].slide(0,0,keyIncrement);    break;}
    }
  break;
  } // end move out
  case Qt::Key_I :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[0].moveEye(0,0,-keyIncrement);  break;}
      case MOVELOOK :  { m_cameras[0].moveLook(0,0,-keyIncrement); break;}
      case MOVEBOTH :  { m_cameras[0].moveBoth(0,0,-keyIncrement); break;}
      case MOVESLIDE : { m_cameras[0].slide(0,0,-keyIncrement);    break;}
    }
  break;
  } // end move in
  default : break;
  }
  // finally update the GLWindow and re-draw
 updateGL();
}

void GLWindow::updateNearFar(FARNEAR _which)
{
	if(_which ==DECNEAR)
	{

		if(--m_near==0.0)
		{
			m_near=1.0;
		}
	}
	else if(_which == INCNEAR)
	{

		if(++m_near > m_far)
		{
			m_near=m_far-1;
		}
	}
	else if(_which == DECFAR)
	{
		if(--m_far < m_near)
		{
			m_far=m_near+1;
		}
	}
	else if(_which == INCFAR)
	{
		++m_far;
	}
setCameraShape();
}
